<?php
/**
 * Copyright (C) 2025  Jaap Jansma (jaap.jansma@civicoop.org)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

namespace Civi\DataProcessor\FieldOutputHandler;

use Civi\DataProcessor\DataSpecification\FieldSpecification;
use Civi\DataProcessor\Source\SourceInterface;
use CRM_Dataprocessor_ExtensionUtil as E;

class FallbackFieldOutputHandler extends AbstractFieldOutputHandler {

  /**
   * @var \Civi\DataProcessor\DataSpecification\FieldSpecification
   */
  protected $outputFieldSpec;

  /**
   * @var String
   */
  protected $field;

  /**
   * @var String
   */
  protected $fallbackField;

  /**
   * @var string
   */
  protected $what;

  /**
   * @var SourceInterface
   */
  protected $dataSource;

  /**
   * @var \Civi\DataProcessor\DataSpecification\FieldSpecification
   */
  protected $inputFieldSpec;

  /**
   * @var SourceInterface
   */
  protected $dataSourceFallback;

  /**
   * @var \Civi\DataProcessor\DataSpecification\FieldSpecification
   */
  protected $inputFallbackFieldSpec;

  /**
   * Initialize the processor
   *
   * @param String $alias
   * @param String $title
   * @param array $configuration
   * @param \Civi\DataProcessor\ProcessorType\AbstractProcessorType $processorType
   */
  public function initialize($alias, $title, $configuration) {
    $this->outputFieldSpec = new FieldSpecification($alias, $this->getType(), $title);
    $this->what = $configuration['what'];
    if ($this->what == 'fields') {
      $this->field = $configuration['field'];
      $this->fallbackField = $configuration['fallback_field'];
    } elseif ($this->what == 'datasource') {
      [$this->dataSource, $this->inputFieldSpec] = $this->initializeField($configuration['datasource_field'], $configuration['datasource'], $alias);
      [$this->dataSourceFallback, $this->inputFallbackFieldSpec] = $this->initializeField($configuration['datasource_fallback_field'], $configuration['datasource_fallback'], $alias . '_fallback');
    }
  }

  /**
   * Returns the data type of this field
   *
   * @return String
   */
  protected function getType() {
    return 'String';
  }

  /**
   * Returns the formatted value
   *
   * @param $rawRecord
   * @param $formattedRecord
   *
   * @return \Civi\DataProcessor\FieldOutputHandler\FieldOutput
   */
  public function formatField($rawRecord, $formattedRecord) {
    $rawValue = null;
    $formattedValue = null;
    if ($this->what == 'datasource') {
      $rawValue = $rawRecord[$this->inputFieldSpec->alias] ?? '';
      $formattedValue = $rawRecord[$this->inputFieldSpec->alias] ?? '';
      if ($rawValue == null || $rawValue == '') {
        $rawValue = $rawRecord[$this->inputFallbackFieldSpec->alias] ?? '';
        $formattedValue = $rawRecord[$this->inputFallbackFieldSpec->alias] ?? '';
      }

      if (strpos($formattedValue, \CRM_Core_DAO::VALUE_SEPARATOR) !== false) {
        $formattedValue = explode(\CRM_Core_DAO::VALUE_SEPARATOR, trim($formattedValue,\CRM_Core_DAO::VALUE_SEPARATOR));
        $formattedValue = implode(",", $formattedValue);
      }
    }
    $output = new FieldOutput($rawValue);
    $output->formattedValue = $formattedValue;
    return $output;
  }

  /**
   * @return \Civi\DataProcessor\DataSpecification\FieldSpecification
   */
  public function getOutputFieldSpecification() {
    return $this->outputFieldSpec;
  }

  /**
   * Do post formatting.
   *
   * Post formatting is additional formatting of the formatted values of a
   * record.
   *
   * @param \Civi\DataProcessor\FieldOutputHandler\FieldOutput[] $record
   *
   * @return \Civi\DataProcessor\FieldOutputHandler\FieldOutput[]
   */
  public function postFormat(array $record): array {
    if ($this->what == 'fields') {
      $alias = $this->getOutputFieldSpecification()->alias;
      $field = $record[$this->field];
      $fallback = $record[$this->fallbackField];
      $record[$alias] = clone $fallback;
      if ($field && !$field->isEmpty()) {
        $record[$alias] = clone $field;
      }
    }
    return $record;
  }

  /**
   * Returns true when this handler has additional configuration.
   *
   * @return bool
   */
  public function hasConfiguration() {
    return true;
  }

  /**
   * When this handler has additional configuration you can add
   * the fields on the form with this function.
   *
   * @param \CRM_Core_Form $form
   * @param array $field
   */
  public function buildConfigurationForm(\CRM_Core_Form $form, $field=array()) {
    $dataSourceFields =  \CRM_Dataprocessor_Utils_DataSourceFields::getAvailableFieldsInDataSources($field['data_processor_id']);
    $fieldSelect = $this->getFieldOutputHandlers($field['data_processor_id']);

    $whatOptions = [
      'datasource' => E::ts('Select fields from a data source'),
      'fields' => E::ts('Select fields from field output'),
    ];

    $form->add('select', 'what', E::ts('What'), $whatOptions, true, array(
      'style' => 'min-width:250px',
      'class' => 'crm-select2 huge',
      'placeholder' => E::ts('- select -'),
    ));

    $form->add('select', 'field', E::ts('Field'), $fieldSelect, false, array(
      'style' => 'min-width:250px',
      'class' => 'crm-select2 huge',
      'placeholder' => E::ts('- select -'),
    ));
    $form->add('select', 'fallback_field', E::ts('Fallback field'), $fieldSelect, false, array(
      'style' => 'min-width:250px',
      'class' => 'crm-select2 huge',
      'placeholder' => E::ts('- select -'),
    ));

    $form->add('select', 'datasource_field', E::ts('Field'), $dataSourceFields, false, array(
      'style' => 'min-width:250px',
      'class' => 'crm-select2 huge',
      'placeholder' => E::ts('- select -'),
    ));
    $form->add('select', 'datasource_fallback_field', E::ts('Fallback field'), $dataSourceFields, false, array(
      'style' => 'min-width:250px',
      'class' => 'crm-select2 huge',
      'placeholder' => E::ts('- select -'),
    ));

    if (isset($field['configuration'])) {
      $configuration = $field['configuration'];
      $defaults = array();
      if (isset($configuration['field'])) {
        $defaults['field'] = $configuration['field'];
      }
      if (isset($configuration['fallback_field'])) {
        $defaults['fallback_field'] = $configuration['fallback_field'];
      }
      if (isset($configuration['what'])) {
        $defaults['what'] = $configuration['what'];
      } else {
        $defaults['what'] = 'fields';
      }
      if (isset($configuration['datasource_field']) && isset($configuration['datasource'])) {
        $defaults['datasource_field'] = \CRM_Dataprocessor_Utils_DataSourceFields::getSelectedFieldValue($field['data_processor_id'], $configuration['datasource'], $configuration['datasource_field']);
      }
      if (isset($configuration['datasource_fallback_field']) && isset($configuration['datasource_fallback'])) {
        $defaults['datasource_fallback_field'] = \CRM_Dataprocessor_Utils_DataSourceFields::getSelectedFieldValue($field['data_processor_id'], $configuration['datasource_fallback'], $configuration['datasource_fallback_field']);
      }
      $form->setDefaults($defaults);
    }
    $form->addFormRule([FallbackFieldOutputHandler::class, 'validateConfigurationForm']);
  }

  public static function validateConfigurationForm(array $values) {
    $errors = [];
    if ($values['what'] == 'fields') {
      if (empty($values['field'])) {
        $errors['field'] = E::ts('The field is required.');
      }
      if (empty($values['fallback_field'])) {
        $errors['fallback_field'] = E::ts('The field is required.');
      }
    } elseif ($values['what'] == 'datasource') {
      if (empty($values['datasource_field'])) {
        $errors['datasource_field'] = E::ts('The field is required.');
      }
      if (empty($values['datasource_fallback_field'])) {
        $errors['datasource_fallback_field'] = E::ts('The field is required.');
      }
    }
    return empty($errors) ? TRUE : $errors;
  }

  /**
   * When this handler has configuration specify the template file name
   * for the configuration form.
   *
   * @return false|string
   */
  public function getConfigurationTemplateFileName() {
    return "CRM/Dataprocessor/Form/Field/Configuration/FallbackFieldOutputHandler.tpl";
  }


  /**
   * Process the submitted values and create a configuration array
   *
   * @param $submittedValues
   * @return array
   */
  public function processConfiguration($submittedValues) {
    $configuration['what'] = $submittedValues['what'];
    if ($configuration['what'] == 'fields') {
      $configuration['field'] = $submittedValues['field'];
      $configuration['fallback_field'] = $submittedValues['fallback_field'];
    } elseif ($configuration['what'] == 'datasource') {
      [$datasource, $field] = explode('::', $submittedValues['datasource_field'], 2);
      $configuration['datasource_field'] = $field;
      $configuration['datasource'] = $datasource;
      [$datasource_fallback, $fallback_field] = explode('::', $submittedValues['datasource_fallback_field'], 2);
      $configuration['datasource_fallback_field'] = $fallback_field;
      $configuration['datasource_fallback'] = $datasource_fallback;
    }
    return $configuration;
  }

  protected function getFieldOutputHandlers(int $dataProcessorId) {
    $return = [];
    $dataProcessor = civicrm_api3('DataProcessor', 'getsingle', array('id' => $dataProcessorId));
    $dataProcessorClass = \CRM_Dataprocessor_BAO_DataProcessor::dataProcessorToClass($dataProcessor);
    $outputFieldHandlers = $dataProcessorClass->getDataFlow()->getOutputFieldHandlers(FALSE);
    foreach ($outputFieldHandlers as $outputFieldHandler) {
      if ($outputFieldHandler->getOutputFieldSpecification()->alias != $this->getOutputFieldSpecification()->alias) {
        $return[$outputFieldHandler->getOutputFieldSpecification()->alias] = $outputFieldHandler->getOutputFieldSpecification()->title;
      }
    }
    return $return;
  }

}
